from py4j.java_gateway import get_field

from antlr4 import InputStream, CommonTokenStream
from .grammar import NeuralogicLexer, NeuralogicParser
from .error import InvalidRuleException


class Rule:
    @staticmethod
    def from_str(rule: str) -> "Rule":
        lexer = NeuralogicLexer(InputStream(rule))
        stream = CommonTokenStream(lexer)
        parser = NeuralogicParser(stream)
        parser.examplesFile()

        if parser.getNumberOfSyntaxErrors() != 0:
            raise InvalidRuleException

        return Rule(rule)

    def __init__(self, rule: str):
        self.original_string = rule

    @property
    def weight(self):
        return get_field(self.rule, "weight")

    @weight.setter
    def weight(self, x):
        pass

    @property
    def offset(self):
        return get_field(self.rule, "offset")

    @offset.setter
    def offset(self, x):
        pass

    @property
    def head(self):
        return get_field(self.rule, "head")

    @head.setter
    def head(self, x):
        pass

    @property
    def aggregation_function(self):
        return get_field(self.rule, "aggregationFcn")

    @aggregation_function.setter
    def aggregation_function(self, x):
        pass

    @property
    def activation_function(self):
        return get_field(self.rule, "activationFcn")

    @activation_function.setter
    def activation_function(self, x):
        pass

    @property
    def metadata(self):
        return get_field(self.rule, "metadata")

    @metadata.setter
    def metadata(self, x):
        pass



from . import get_neuralogic
from py4j.java_gateway import get_field
from .settings import Settings
from .sources import Sources


class Pipeline:
    def __init__(self, settings: Settings, sources: Sources):
        self.namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.building
        self.pipeline = self.namespace.LearningSchemeBuilder.getPipeline(settings.settings, sources.sources)

    def execute(self, sources: Sources):
        result = self.pipeline.execute(sources.sources)

        name = get_field(result, "r")
        data = get_field(result, "s")

        return name, data



from typing import Union
from pathlib import Path
import sys
import os
from py4j.java_gateway import JavaGateway

neuralogic = None
gateway = None


def set_environment(java_home: Union[Path, str], class_path: Union[Path, str]):
    """
    Set class path and java home environment variables

    :param java_home:
    :param class_path:
    :return:
    """
    os.environ["JAVA_HOME"] = java_home
    os.environ["CLASS_PATH"] = class_path


def initialize(path: Union[Path, str], std_out=sys.stdout, std_err=sys.stderr, die_on_exit=True):
    global gateway, neuralogic

    gateway = JavaGateway.launch_gateway(
        classpath=path,
        redirect_stdout=std_out,
        redirect_stderr=std_err,
        die_on_exit=die_on_exit,
    )
    neuralogic = gateway.jvm


def get_gateway():
    return gateway


def get_neuralogic():
    return neuralogic



from .builder import Weight, Sample
from typing import List
import torch
from torch import Tensor
from torch.nn import Parameter, ParameterList
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.inits import zeros


def already_seen(seen, neurons):
    for neuron in neurons:
        if not seen[neuron]:
            return False
    return True


def to_layers(sample: Sample):
    seen = [False] * len(sample.neurons)
    layers = []

    while True:
        index = 0
        current_layer = []

        for i in range(index, len(sample.neurons)):
            neuron = sample.neurons[i]
            if seen[neuron.index]:
                continue
            if not already_seen(seen, neuron.inputs):
                if index == 0:
                    index = i
                continue
            current_layer.append(neuron)
        layers.append(current_layer)
        for n in current_layer:
            seen[n.index] = True
        if index == 0:
            return layers


class NeuraLogicHelperLayer(MessagePassing):
    def __init__(self, activation, weights):
        super(NeuraLogicHelperLayer, self).__init__(aggr=NeuraLogicLayer.aggregations.get(activation, "add"))

        self.activation = NeuraLogicLayer.activations.get(activation, None)
        self.weights = weights

    def forward(self, layer_input):
        x, neurons = layer_input

        from_neuron = []
        to_neuron = []
        edge_weight = []

        for neuron in neurons:
            if neuron.weights is None:
                edge_weight = None
                for input in neuron.inputs:
                    from_neuron.append(input)
                    to_neuron.append(neuron.index)
            else:
                for input, weight in zip(neuron.inputs, neuron.weights):
                    from_neuron.append(input)
                    to_neuron.append(neuron.index)
                    edge_weight.append(self.weights[weight + 1])

        if len(from_neuron) == 0:
            return x

        edge_index = torch.tensor([from_neuron, to_neuron])
        out = self.propagate(edge_index, x=x, edge_weight=edge_weight)

        if self.activation:
            out = self.activation(out)

        return x + out

    def message(self, x_j: Tensor, edge_weight=None) -> Tensor:
        out = x_j

        if edge_weight is None:
            return out

        output = []

        for x, w in zip(out, edge_weight):
            value = x * w
            output.append(value)

        return torch.stack(output)


class NeuraLogicLayer(torch.nn.Module):
    activations = {
        "Sigmoid": torch.sigmoid,
        "ReLu": torch.relu,
        "Tanh": torch.tanh,
    }

    aggregations = {
        "Average": "mean",
        "Maximum": "max",
    }

    def __init__(self, weights: List[Weight]):
        super(NeuraLogicLayer, self).__init__()

        params = [
            Parameter(torch.tensor(weight.dimensions, dtype=torch.float), requires_grad=not weight.fixed)
            for weight in weights
        ]

        self.built_layers = False
        self.neura_layers: List[NeuraLogicHelperLayer] = []
        self.weights = ParameterList(params)

        self.reset_params()

    def reset_params(self):
        self.weights[0][0] = 1

        for weight in self.weights[1:]:
            zeros(weight)

    def forward(self, sample: Sample):
        layers = to_layers(sample)

        if not self.built_layers:
            self.built_layers = True
            self.neura_layers = [NeuraLogicHelperLayer(layer[0].activation, self.weights) for layer in layers]

        x = torch.zeros((len(sample.neurons), 1))
        x[layers[0][0].index] = 1

        for n_layer, layer in zip(self.neura_layers, layers):
            x = n_layer((x, layer))
        return x[layers[-1][0].index]



from typing import List
from .builder import Sample, Weight, Neuron, Model, OFFSET
import dynet as dy


class DyNetDeserializer:
    activations = {
        "Sigmoid": dy.logistic,
        "Average": dy.average,
        "Maximum": dy.emax,
        "ReLu": dy.rectify,
        "Tanh": dy.tanh,
    }

    def __init__(self, model: Model):
        self.model = dy.ParameterCollection()
        self.weights = self.deserialize_weights(model.weights)

    def deserialize_weights(self, weights: List[Weight]):
        return [weight.value if weight.fixed else self.model.add_parameters(weight.dimensions) for weight in weights]

    def build_neuron_expr(self, neuron: Neuron, neuron_expressions: List):
        return DyNetDeserializer.to_dynet_expression(neuron, neuron_expressions, self.weights)

    def build_sample(self, sample: Sample):
        dynet_neurons = [None] * len(sample.neurons)

        for neuron in sample.neurons:
            dynet_neurons[neuron.index] = self.build_neuron_expr(neuron, dynet_neurons)
        return dynet_neurons[sample.neurons[-1].index]

    @staticmethod
    def to_dynet_value(neuron: Neuron):
        dim = 1

        if hasattr(neuron.value, "__len__"):
            dim = len(neuron.value)
        if dim == 1:
            return dy.scalarInput(float(neuron.value))
        if dim > 1:
            res = dy.vecInput(dim)
            res.set(neuron.value)
            return res
        return dy.inputTensor(neuron.value, dim)

    @staticmethod
    def to_dynet_expression(neuron: Neuron, neurons: List, weights: List):
        out = None

        if neuron.inputs:
            if neuron.weights:
                for w, i in zip(neuron.weights, neuron.inputs):
                    if not out:
                        out = weights[w + OFFSET] * neurons[i]
                    else:
                        out += weights[w + OFFSET] * neurons[i]
            else:
                for i in neuron.inputs:
                    if not out:
                        out = neurons[i]
                    else:
                        out += neurons[i]
        else:
            out = DyNetDeserializer.to_dynet_value(neuron)

        if neuron.offset:
            out += weights[neuron.offset + OFFSET]

        if neuron.activation:
            if neuron.pooling:
                out = list(out)
            return DyNetDeserializer.activations[neuron.activation](out)
        return out



from . import get_neuralogic
from .helpers import stream_to_list
from .settings import Settings
from .sources import Sources
import json
from typing import List
from py4j.java_gateway import get_field


OFFSET = 1


class Sample:
    def __init__(self, sample):
        self.id = get_field(sample, "id")
        self.target = float(get_field(sample, "target"))
        self.output_neuron = get_field(sample, "neuron")
        self.neurons = self.deserialize_network(get_field(sample, "network"))
        self.output_neuron = self.neurons[-1].index

    def deserialize_network(self, network):
        neurons = []

        for i, neuron in enumerate(network):
            neuron_object = Neuron(neuron, i)
            neurons.append(neuron_object)

            if neuron_object.index == self.output_neuron:
                break
        return neurons


class Neuron:
    def __init__(self, neuron, index):
        self.index = index
        self.name = get_field(neuron, "name")

        self.weighted = get_field(neuron, "weighted")

        self.activation = get_field(neuron, "activation")
        self.inputs = get_field(neuron, "inputs")

        self.weights = get_field(neuron, "weights")
        self.offset = get_field(neuron, "offset")

        self.value = get_field(neuron, "value")
        self.pooling = get_field(neuron, "pooling")


class Weight(object):
    def __init__(self, weight):
        self.index = get_field(weight, "index")
        self.name = get_field(weight, "name")
        self.dimensions = tuple(get_field(weight, "dimensions"))
        self.value = json.loads(get_field(weight, "value"))
        self.fixed = get_field(weight, "isFixed")

        if not isinstance(self.value, list):
            self.value = [self.value]

        if not self.dimensions:
            self.dimensions = (1,)


class Model:
    def __init__(self, weights: List[Weight], samples: List[Sample]):
        self.weights = weights
        self.samples = samples

    @staticmethod
    def from_neuralogic(settings: Settings, sources: Sources) -> "Model":
        namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.building
        pipes_namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.pipes.specific

        builder = namespace.End2endTrainigBuilder(settings.settings, sources.sources)
        nn_builder = builder.getEnd2endNNBuilder()

        pipeline = nn_builder.buildPipeline()
        serializer_pipe = pipes_namespace.NeuralSerializerPipe()

        pipeline.connectAfter(serializer_pipe)
        pipeline.execute(sources.sources)
        result = serializer_pipe.get()

        weights = sorted([Weight(x) for x in get_field(result, "r")], key=lambda a: a.index)
        sample = [Sample(x) for x in stream_to_list(get_field(result, "s"))]

        return Model(weights, sample)



from . import get_neuralogic
from py4j.java_gateway import get_field


class Settings:
    def __init__(self):
        self.namespace = get_neuralogic().cz.cvut.fel.ida.setup
        self.settings = self.namespace.Settings.forFastTest()

    @property
    def seed(self) -> int:
        return self.settings.seed

    @seed.setter
    def seed(self, seed: int):
        self.settings.seed = seed

    @property
    def default_fact_value(self) -> float:
        return self.settings.defaultFactValue

    @default_fact_value.setter
    def default_fact_value(self, value: float):
        self.settings.defaultFactValue = value

    def to_json(self) -> str:
        return self.settings.exportToJson()



from . import get_neuralogic, get_gateway
from .settings import Settings
from .error import InvalidLearningModeException

from enum import Enum

from typing import Union
from pathlib import Path


class LearningMode(Enum):
    CROSSVALIDATION = 0
    TRAIN_TEST = 1
    TRAIN_ONLY = 2
    TEST_ONLY = 3


class Sources:
    @staticmethod
    def from_dir(source_dir: Union[Path, str], settings: Settings) -> "Sources":
        neuralogic = get_neuralogic()
        gateway = get_gateway()

        args = ["-sd", str(source_dir), "-ts", "100"]
        jargs = gateway.new_array(gateway.jvm.java.lang.String, len(args))

        for i, item in enumerate(args):
            jargs[i] = item

        sources = neuralogic.cz.cvut.fel.ida.neuralogic.cli.utils.Runner.getSources(jargs, settings.settings)
        return Sources(sources)

    @staticmethod
    def from_str(text: str, settings: Settings, mode: LearningMode) -> "Sources":
        neuralogic = get_neuralogic()
        reader = neuralogic.java.io.StringReader(text)

        sources = neuralogic.cz.cvut.fel.ida.setup.Sources(settings.settings)
        # sources.setTemplateReader(reader)

        if mode == LearningMode.CROSSVALIDATION:
            sources.crossvalidation = True
        elif mode == LearningMode.TRAIN_TEST:
            sources.trainTest = True
        elif mode == LearningMode.TRAIN_ONLY:
            sources.trainOnly = True
        elif mode == LearningMode.TEST_ONLY:
            sources.testOnly = True
        else:
            raise InvalidLearningModeException()
        return Sources(sources)

    def __init__(self, sources):
        self.sources = sources

    def to_json(self) -> str:
        return self.sources.exportToJson()

    # def __init__(self, settings: Settings, source_dir: Union[Path, str]):
    #     neuralogic = get_neuralogic()
    #     gateway = get_gateway()
    #
    #     args = ["-sd", str(source_dir), "-ts", "100"]
    #     jargs = gateway.new_array(gateway.jvm.java.lang.String, len(args))
    #
    #     for i, item in enumerate(args):
    #         jargs[i] = item
    #
    #     self.sources = (
    #         neuralogic.cz.cvut.fel.ida.neuralogic.cli.utils.Runner.getSources(
    #             jargs, settings.settings
    #         )
    #     )
